#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

require 'ftools'
require 'langscan'

module Gonzui
  include Util

  class Deindexer
    def initialize (config, dbm, normalized_path)
      @config = config
      @dbm = dbm
      @normalized_path = normalized_path
      @path_id = @dbm.get_path_id(@normalized_path)
      @package_id = @dbm.get_package_id_from_path_id(@path_id)
      @package_name = @dbm.get_package_name(@package_id)
    end

    private
    def remove_words_of_source
      word_ids  = @dbm.get_word_ids(@path_id)
      word_ids.each {|word_id|
        word_package_id = word_id + @package_id
        @dbm.wordpkgid_pathids.delete_both(word_package_id, @path_id)
        unless @dbm.wordpkgid_pathids.has_key?(word_package_id)
          @dbm.wordid_pkgids.delete_both(word_id, @package_id)
        end

        unless @dbm.wordid_pkgids.has_key?(word_id)
          word = @dbm.wordid_word[word_id]
          assert_non_nil(word)
          @dbm.wordid_word.delete(word_id) 
          @dbm.word_wordid.delete(word)
          @dbm.decrease_counter(:nwords)
        end

        word_path_id = word_id + @path_id
        @dbm.wordpathid_info.delete(word_path_id)
      }
      @dbm.pathid_wordids.delete(@path_id)
    end

    def remove_digest
      lang_id = @dbm.get_lang_id_from_path_id(@path_id)
      digest = @dbm.get_digest(@path_id)

      lang_abbrev = @dbm.get_lang_abbrev(lang_id)
      content = @dbm.get_file_content(@path_id)
      if lang_abbrev
        funcfreq = @dbm.make_funcfreq(lang_abbrev)
        digest.each {|info|
          if Type.function?(info.type)
            function = content[info.byteno, info.length]
            funcfreq.decrease(function)
          end
        }
      end
      @dbm.pathid_digest.delete(@path_id)
    end

    def remove_path
      path = @dbm.get_path(@path_id)
      @dbm.path_pathid.delete(@normalized_path)
      @dbm.pathid_path.delete(@path_id)
      @dbm.pkgid_pathids.delete_both(@package_id, @path_id)
    end

    def remove_language
      lang_id = @dbm.get_lang_id_from_path_id(@path_id)
      lang_abbrev = @dbm.get_lang_abbrev(lang_id)
      @dbm.decrease_counter(@dbm.make_nsources_key(lang_abbrev))
      nsources = @dbm.get_nsources_by_lang_id(lang_id)
      if nsources == 0
        lang_name = @dbm.get_lang_name(lang_id)
        @dbm.langid_lang.delete(lang_id)
        @dbm.langid_labbr.delete(lang_id)
        @dbm.labbr_langid.delete(lang_abbrev)
        @dbm.decrease_counter(:nlanguages)
      end
    end

    def remove_source_if_necessary
      return unless @dbm.source_file?(@path_id)
      remove_digest
      remove_words_of_source
      remove_language
      nlines = @dbm.get_file_info(@path_id).nlines
      @dbm.decrease_counter(:nsources)
      @dbm.decrease_counter(:nlines, nlines)
    end

    def remove_package_if_necessary
      return unless package_empty?
      @dbm.pkg_pkgid.delete(@package_name)
      @dbm.pkgid_pkg.delete(@package_id)
      @dbm.pkgid_langids.delete(@package_id)
      @dbm.decrease_counter(:npackages)
    end

    def remove_file
      @dbm.pathid_content.delete(@path_id)
      @dbm.pathid_pkgid.delete(@path_id)
      @dbm.pathid_info.delete(@path_id)
      @dbm.decrease_counter(:nfiles)
    end

    def package_empty?
      @dbm.get_path_ids(@package_id).empty?
    end

    public 
    def deindex
      remove_source_if_necessary
      remove_path
      remove_file
      remove_package_if_necessary
    end
  end
end
