#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

module Gonzui
  class Package
    def initialize (directory)
      @name      = File.basename(directory)
      @directory = directory
      @relative_paths   = []
      import_files
    end
    attr_reader :name
    attr_reader :directory

    def import_files
      raise "directory or name is unset" if @directory.nil? or @name.nil?
      Dir.all_files(@directory).each {|file_name|
        relative_path = File.relative_path(file_name, @directory)
        @relative_paths.push(relative_path)
      }
    end

    public
    def include? (relative_path)
      @relative_paths.include?(relative_path)
    end

    def get_path_id (relative_path)
      @relative_paths.index(relative_path)
    end

    def each
      @relative_paths.each {|relative_path|
        real_path = File.join(@directory, relative_path)
        normalized_path = File.join(@name, relative_path)
        yield(real_path, normalized_path)
      }
    end

    def nfiles
      @relative_paths.length
    end
  end
end
