#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

module Gonzui
  class SearchServlet < GonzuiAbstractServlet
    def self.mount_point
      "/search"
    end

    def do_make_status_line (nhits)
      right = ""
      if nhits > 0
        qs = @search_query.string
        from = (@search_query.package or @search_query.path)
        qs = @search_query.simplified_string if from
        to = [nhits, @from + @nresults_per_page].min
        right = [:span]
        right.push(_("Results "), 
                   [:strong, commify(@from + 1)], " - ", 
                   [:strong, commify(to)], _(" of "), 
                   [:strong, commify(nhits)], _(" for "),
                   [:strong, qs])
        right.push(_(" from "), [:strong, from]) if from
      end
      center =  [:a, {:href => make_google_url(@search_query)},
        _("Google it")]
      return center, right
    end

    def make_emphasized_line (content, lineno, line_range, *occ_list)
      line = [:span]
      occ_list.each {|occ|
        occ_range = (occ.byteno ... (occ.byteno + occ.length))
        pre = range = nil
        if line_range.include?(occ.byteno)
          pre = content.substring(line_range.first...occ.byteno)
          range = (occ.byteno ... [occ_range.last, line_range.last].min)
        elsif occ_range.include?(line_range.first)
          range = (line_range.first ... [occ_range.last, line_range.last].min)
        end
        line.push(pre) if pre
        if range
          text = content[range]
          line.push([:strong, text])
          line_range = (range.last ... line_range.last)
        end
      }
      remaining = content.substring(line_range)
      line.push(remaining)
      return line
    end

    def make_snippet_kwic_single (content, item)
      lines_with_lineno = []
      occ = item.list.first
      text = content[occ.byteno, occ.length]
      content.each_line_range(occ.byteno, @ncontexts) {|lineno_offset, range|
        lineno = occ.lineno + lineno_offset
        line = make_emphasized_line(content, lineno, range, occ)
        lines_with_lineno.push([line, lineno])
      }
      return lines_with_lineno
    end

    def make_occurrences_by_lineno (item)
      occurrences_by_lineno = Hash.new
      item.list.sort_by {|occ| occ.byteno}.each {|occ|
        occurrences_by_lineno[occ.lineno] ||= []
        occurrences_by_lineno[occ.lineno].push(occ)
      }
      return occurrences_by_lineno
    end

    def add_contexts (content, lines_with_lineno, occ)
      content.each_line_range(occ.byteno, @ncontexts) {|lineno_offset, range|
        lineno = occ.lineno + lineno_offset
        line = content.substring(range)
        lines_with_lineno.push([line, lineno]) unless lineno == occ.lineno
      }
      return lines_with_lineno.sort_by {|line, lineno| lineno }
    end

    def make_snippet_kwic_multi (content, item)
      seen = Hash.new
      lines_with_lineno = []
      occurrences_by_lineno = make_occurrences_by_lineno(item)
      occurrences_by_lineno.keys.sort.each {|lineno|
        occ_list = occurrences_by_lineno[lineno]
        range = content.line_range(occ_list.first.byteno)
        line = make_emphasized_line(content, lineno, range, *occ_list)
        lines_with_lineno.push([line, lineno])
      }
      if occurrences_by_lineno.length == 1
        occ = item.list.first
        lines_with_lineno = add_contexts(content, lines_with_lineno, occ)
      end
      return lines_with_lineno
    end

    def make_snippet_kwic (content, item)
      if @search_query.length == 1
        return make_snippet_kwic_single(content, item)
      else
        return make_snippet_kwic_multi(content, item)
      end
    end

    def make_snippet (item, markup_url)
      div = [:div]
      content = @dbm.get_file_content(item.path_id)
      lines_with_lineno = make_snippet_kwic(content, item)
      pre = [:pre, {:class => "lines"}]
      lines_with_lineno.sort_by {|line, lineno|
        lineno
      }.each {|line, lineno|
        lineno_url = make_lineno_url(markup_url, lineno)
        lineno_mark = [:a, {:href => lineno_url}, lineno.to_s + ": "]
        pre.push(lineno_mark, line)
        pre.push("\n")
      }
      div.push([:dd, pre])
    end

    def make_expansion_link (item)
      if item.has_more_in_package?
        package_name = @dbm.get_package_name(item.package_id)
        qs = sprintf("package:%s %s", package_name, 
                     @search_query.simplified_string)
        search_query = SearchQuery.new(@config, qs)
        url = make_search_url(search_query)
        from = package_name
      elsif item.has_more_in_path?
        path = @dbm.get_path(item.path_id)
        qs = sprintf("path:%s %s", path, 
                     @search_query.simplified_string)
        search_query = SearchQuery.new(@config, qs)
        url = make_search_url(search_query)
        from = File.basename(path)
      else
        assert_not_reached
      end
      dd = [:dd,[:a, {:href => url, :class => "more" }, 
          sprintf(_("More results from %s"), from)]]
      return dd
    end

    def make_package_name_line (package_name)
      url = make_markup_url(package_name)
      dt = [:dt, {:class => "package"}, [:a, {:href => url}, package_name]]
      return dt
    end

    def make_source_path_line (item, package_name, path, markup_url)
      lineno_url = make_lineno_url(markup_url, item.list.first.lineno)
      source_line = [:dt, {:class => "source"}, [:a, 
          {:href => lineno_url}, 
          File.relative_path(path, package_name)]]
      file_info = @dbm.get_file_info(item.path_id)
      size = [:span, {:class => "size"}, 
        sprintf(_(" - %s - %s lines"), 
                format_bytes(file_info.size),
                commify(file_info.nlines))
              ]
      source_line.push(size)
    end

    def make_result_list_by_snippet (result)
      dl = [:dl]
      prev_package_id = nil
      nresults = 0

      result.each_from(@from) {|item|
        assert(item.list.length > 0)
        package_name = @dbm.get_package_name(item.package_id)

        if item.package_id != prev_package_id
          dt = make_package_name_line(package_name)
          dl.push(dt)
          prev_package_id = item.package_id
        end

        path = @dbm.get_path(item.path_id)
        markup_url = make_markup_url(path, @search_query)
        dt = make_source_path_line(item, package_name, path, markup_url)
        dl.push(dt)

        snippet = make_snippet(item, markup_url)
        dl.push(snippet)
        if item.has_more?
          link = make_expansion_link(item)
          dl.push(link)
        end
        nresults += 1
        break if nresults >= @nresults_per_page
      }
      return dl
    end

    def make_separator
      [:div, {:class => "separator"}, ""]
    end

    def collect_context_lines (item)
      content = @dbm.get_file_content(item.path_id)
      context_lines = Hash.new

      occurrences_by_lineno = make_occurrences_by_lineno(item)
      occurrences_by_lineno.keys.sort.each {|lineno|
        occ_list = occurrences_by_lineno[lineno]
        range = content.line_range(occ_list.first.byteno)
        occ = occ_list.first
        content.each_line_range(occ.byteno, @ncontexts) {|lineno_offset,range|
          lineno = occ.lineno + lineno_offset
          next if context_lines.has_key?(lineno) and lineno_offset != 0
          line = make_emphasized_line(content, lineno, range, *occ_list)
          context_lines[lineno] = line
        }
      }
      return context_lines
    end

    def make_result_list_like_grep (result)
      assert_equal(1, result.length)
      dl = [:dl]
      item = result.first

      package_name = @dbm.get_package_name(item.package_id)
      path = @dbm.get_path(item.path_id)
      markup_url = make_markup_url(path, @search_query)

      dl.push(make_package_name_line(package_name))
      dl.push(make_source_path_line(item, package_name, path, markup_url))

      context_lines = collect_context_lines(item)

      pre = [:pre, {:class => "lines"}]
      prev_lineno = nil
      context_lines.keys.sort.each {|lineno|
        lineno_url = make_lineno_url(markup_url, lineno)
        lineno_mark = [:a, {:href => lineno_url }, lineno.to_s + ": "]
        pre.push(make_separator)if prev_lineno and lineno > prev_lineno + 1
        pre.push(lineno_mark, context_lines[lineno], "\n")
        prev_lineno = lineno
      }
      dl.push([:dd, pre])
      return dl
    end

    def do_make_navi (from)
      make_search_url(@search_query, 
                      :from => from,
                      :nresults_per_page => @nresults_per_page)
    end

    def redirect_to_url (url)
      @response.set_redirect(HTTPStatus::MovedPermanently, url)
      assert_not_reached
    end

    # FIXME: it's not used 
    def redirect_to_single_result (result)
      item = result.first
      occ = item.list.first
      path = @dbm.get_path(item.path_id)
      lineno = (occ.lineno or 1)
      markup_url = make_markup_url(path, @search_query)
      lineno_url = make_lineno_url(markup_url, lineno)
      redirect_to_url(lineno_url)
    end


    def make_not_found_message
      message = [:div, {:class => "message"},
        sprintf(_("No sources were found matching your search - %s."), 
                @search_query.string)
      ]
      return message
    end

    def make_query_error_message (e)
      message =[:div]
      message.push(make_not_found_message)
      message.push([:div, {:class => "message"}, _(e.message)])
      return message
    end

    def make_result_content (result)
      @from = 0 if @from >= result.length

      if result.empty?
        make_not_found_message
      elsif result.single_path?
        make_result_list_like_grep(result)
      else
        make_result_list_by_snippet(result)
      end
    end

    # FIXME: it's not used 
    def redirect_if_neccessary
      if @search_query.empty?
        redirect_to_url("/")
      elsif @search_query.path_only?
        markup_url = make_markup_url(@search_query.path, @search_query)
        redirect_to_url(markup_url)
      elsif @search_query.package_only?
        markup_url = make_markup_url(@search_query.package, @search_query)
        redirect_to_url(markup_url)
      end
    end

    def make_result_navi (result)
      make_navi(result.length, @from, @nresults_per_page, @config.max_pages)
    end

    def make_notice
      if @search_query.ignored_words.empty?
        return ""
      else
        div = [:div, {:class => "notice"}]
        message = 
          sprintf(_("\"%s\" (and any subsequent words) was ignored because queries are limited to %d words."),
                  @search_query.ignored_words.first,
                  @config.max_words
                  )
        div.push(message)
        return div
      end
    end

    def do_GET (request, response)
      init_servlet(request, response)
      log(@search_query.string)
      @ncontexts = 1

      title = make_title(@search_query.string)
      html = make_html
      head = [:head, [:title, title],  make_script, *make_meta_and_css]
      body = [:body]
      body.push(make_h1)
      body.push(make_search_form)

      nhits = 0
      navi = nil
      begin
        @search_query.validate
        redirect_if_neccessary

        searcher = Searcher.new(@dbm, @search_query, 
                                @config.max_results_overall)
        result = searcher.search
        content = make_result_content(result)
        if result.length > @nresults_per_page
          navi = make_result_navi(result)
        end
        nhits = result.length
      rescue QueryError => e
        content = make_query_error_message(e)
      end
      status_line = make_status_line(_("Search"), nhits)

      body.push(make_notice)
      body.push(status_line)
      body.push(content)
      body.push(navi) if navi

      body.push(make_footer)
      html.push(head)
      html.push(body)
      set_content_type_text_html
      response.body = format_html(html)
    end
  end

  GonzuiServlet.register(SearchServlet)
end
