#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

module Gonzui
  class StatisticsServlet < GonzuiAbstractServlet
    def self.mount_point
      "/stat"
    end

    RankingMax = 100

    def choose_tr_class (i)
      if (i + 1) % 2 == 0 then "even" else "odd" end
    end

    def make_heading (*items)
      tr = [:tr, {:class => "heading"}]
      items.each {|string, klass, colspan|
        td = [:td, {:class => klass}, string]
        td[1][:colspan] = colspan if colspan
        tr.push(td)
      }
      return tr
    end

    def make_bar (freq, max)
      percentage = freq * 100 / max
      style = "width: #{percentage}px;"
      bar = [:div, {:class => "bar", :style => style}, ""]
    end

    def make_function_popularity_table (lang_abbrev)
      table = [:table, {:class => "fullwidth"}]
      heading = make_heading([_("Rank"), "first-center"],
                             [_("Function Name"), "nonfirst"],
                             [_("Frequency"), "nonfirst-center", 2])
      table.push(heading)

      functions = @dbm.get_function_popularity(lang_abbrev, RankingMax)
      max = nil
      functions.each_with_index {|pair, i|
        k = choose_tr_class(i)
        freq = pair.first
        name = pair.last
        max  = freq if max.nil?
        search_query = SearchQuery.new(@config, name, :language => lang_abbrev)
        url  = make_search_url(search_query)
        a = [:a, {:href => url}, name]
        bar = make_bar(freq, max)
        tr = [:tr, {:class => k}, 
          [:td, {:width => "16%", :class => "first-center"}, i + 1],
          [:td, {:width => "60%", :class => "nonfirst"}, a], 
          [:td, {:width => "12%", :class => "nonfirst-right"}, commify(freq)],
          [:td, {:width => "12%"}, bar]
        ]
        table.push(tr)
      }
      return table
    end

    def do_make_status_line
      []
    end

    def make_tr (first, second, third = nil)
      tr = [:tr]
      td1 = [:td, {:class => "first", :width => "76%"}, first]
      if third
        td2 = [:td, {:class => "nonfirst-right", :width => "12%"}, 
          commify(second)]
        td3 = [:td, {:width => "12%"}, third]
      else
        td2 = [:td, {:class => "nonfirst-center", :width => "24%"}, 
          commify(second)]
      end
      tr.push(td1)
      tr.push(td2)
      tr.push(td3) if td3
      return tr
    end

    def make_language_table
      table = [:table, {:class => "fullwidth"}]
      heading = make_heading([_("Source Files by Language"), "first", 3])
      table.push(heading)
      languages = []
      max = -1
      @dbm.each_language {|lang_id, lang_abbrev, lang_name|
        nsources = @dbm.get_nsources_by_lang_id(lang_id)
        languages.push([lang_id, lang_abbrev, lang_name, nsources])
        max = [max, nsources].max
      }
      return if languages.empty?
      languages = languages.sort_by {|id, abbr, name, nsources| - nsources}
      languages.each_with_index {|item, i|
        id, abbr, name, nsources = item
        url = make_stat_url("func/" + abbr)
        a = [:a, {:href => url}, name]
        tr = make_tr(a, nsources, make_bar(nsources, max))
        k = choose_tr_class(i)
        tr.insert(1, {:class => k})
        table.push(tr)
      }
      return table
    end

    def make_overview_table
      table = [:table, {:class => "fullwidth"}]
      heading = make_heading([_("Overview"), "first", 2])
      table.push(heading)
      trs = []
      trs << make_tr(_("Packages"),        @dbm.npackages)
      trs << make_tr(_("Files"),           @dbm.nfiles)
      trs << make_tr(_("Source Files"),    @dbm.nsources)
      trs << make_tr(_("Lines of Code"),   @dbm.nlines)
      trs << make_tr(_("Searchable Keys"), @dbm.nwords)
      trs << make_tr(_("Languages"),       @dbm.nlanguages)
      trs.each_with_index {|tr, i|
        k = choose_tr_class(i)
        tr.insert(1, {:class => k})
        table.push(tr)
      }
      return table
    end

    def make_top_page
      title = make_title(_("Statistics"))
      content = [:div]
      content.push(make_overview_table)
      content.push(make_language_table)
      status_line = make_status_line(_("Statistics"))
      return title, status_line, content
    end

    def make_function_popularity_page
      lang_abbrev = File.basename(@path)
      unless @dbm.has_lang_abbrev?(lang_abbrev)
        raise HTTPStatus::NotFound.new("not found")
      end
      lang_id = @dbm.get_lang_id(lang_abbrev)
      lang_name = @dbm.get_lang_name(lang_id)
      title = make_title(sprintf("statistics/%s", @path))
      url = make_stat_url
      span = [:span,
        [:a, {:href => url}, _("Statistics")], " - ",
        [:a, {:href => ""}, _("Top Functions"), " - ", lang_name]
      ]
      content = make_function_popularity_table(lang_abbrev)
      status_line = make_status_line(span)
      return title, status_line, content
    end

    def do_GET (request, response)
      init_servlet(request, response)
      log()

      @path = make_path
      if @path.empty?
        title, status_line, content = make_top_page
      else
        title, status_line, content = make_function_popularity_page
      end
      
      html = make_html
      head = [:head, [:title, title], *make_meta_and_css]
      body = [:body]
      body.push(make_h1)
      body.push(status_line)
      body.push(content)
      body.push(make_footer)
      html.push(head)
      html.push(body)
      set_content_type_text_html
      response.body = format_html(html)
    end
  end

  GonzuiServlet.register(StatisticsServlet)
end
