module LangScan
  LangScanRegistry = Hash.new

  module_function
  def load_plugins (plugin_path)
    $LOAD_PATH.each {|path|
      candidate_path = File.join(path, plugin_path)
      next unless File.directory?(candidate_path)
      Dir.entries(candidate_path).each {|entry|
        if File.extname(entry) == ".rb" and not /^_/.match(entry)
          begin
            require(File.join(plugin_path, entry))
          rescue LoadError => e
            # ignore load errors
          end
        end
      }
    }
  end

  def load
    load_plugins("langscan")
  end

  def validate_module (mod)
    common_methods = [:lang_name, :lang_abbrev, :scan]
    safe_characters = "[a-z]+"
    common_methods.each {|method|
      raise "#{mod.to_s} lacks #{method}" unless mod.respond_to?(method)
    }
    unless /^#{safe_characters}$/.match(mod.lang_abbrev)
      raise "#{mod.to_s} invalid abbreviation: #{mod.lang_abbrev}"
    end
  end

  def register (mod)
    validate_module(mod)
    mod.extnames.each {|extname|
      LangScanRegistry[extname] = mod
    }
  end

  def modules
    LangScanRegistry.values.uniq
  end

  def choose (file_name)
    extname = File.extname(file_name)
    return LangScanRegistry[extname]
  end

  def support? (file_name)
    extname = File.extname(file_name)
    LangScanRegistry.include?(extname)
  end
end

LangScan.load
